class EventType:
    ASSET_PURCHASE_REQUEST_PROCESSING = 'asset_purchase_request_processing'
    ASSET_CHANGE_REQUEST_PROCESSING = 'asset_change_request_processing'
    ASSET_SUSPEND_REQUEST_PROCESSING = 'asset_suspend_request_processing'
    ASSET_RESUME_REQUEST_PROCESSING = 'asset_resume_request_processing'
    ASSET_CANCEL_REQUEST_PROCESSING = 'asset_cancel_request_processing'
    ASSET_ADJUSTMENT_REQUEST_PROCESSING = 'asset_adjustment_request_processing'
    ASSET_PURCHASE_REQUEST_VALIDATION = 'asset_purchase_request_validation'
    ASSET_CHANGE_REQUEST_VALIDATION = 'asset_change_request_validation'
    PRODUCT_ACTION_EXECUTION = 'product_action_execution'
    PRODUCT_CUSTOM_EVENT_PROCESSING = 'product_custom_event_processing'
    TIER_CONFIG_SETUP_REQUEST_PROCESSING = 'tier_config_setup_request_processing'
    TIER_CONFIG_CHANGE_REQUEST_PROCESSING = 'tier_config_change_request_processing'
    TIER_CONFIG_ADJUSTMENT_REQUEST_PROCESSING = 'tier_config_adjustment_request_processing'
    TIER_CONFIG_SETUP_REQUEST_VALIDATION = 'tier_config_setup_request_validation'
    TIER_CONFIG_CHANGE_REQUEST_VALIDATION = 'tier_config_change_request_validation'
    SCHEDULED_EXECUTION = 'scheduled_execution'
    LISTING_NEW_REQUEST_PROCESSING = 'listing_new_request_processing'
    LISTING_REMOVE_REQUEST_PROCESSING = 'listing_remove_request_processing'
    TIER_ACCOUNT_UPDATE_REQUEST_PROCESSING = 'tier_account_update_request_processing'
    USAGE_FILE_REQUEST_PROCESSING = 'usage_file_request_processing'
    PART_USAGE_FILE_REQUEST_PROCESSING = 'part_usage_file_request_processing'


class TaskCategory:
    BACKGROUND = 'background'
    INTERACTIVE = 'interactive'
    SCHEDULED = 'scheduled'


class ResultType:
    SUCCESS = 'success'
    RESCHEDULE = 'reschedule'
    SKIP = 'skip'
    RETRY = 'retry'
    FAIL = 'fail'



from typing import Literal

from pydantic import BaseModel


class Context(BaseModel):
    installation_id: str
    user_id: str
    account_id: str
    account_role: str
    call_source: Literal['ui', 'api']
    call_type: Literal['admin', 'user']



from connect.eaas.core.enums import ResultType


class _Response:
    def __init__(self, status, output=None):
        self.status = status
        self.output = output

    @classmethod
    def done(cls, *args, **kwargs):
        return cls(ResultType.SUCCESS)

    @classmethod
    def fail(cls, output=None):
        return cls(ResultType.FAIL, output=output)


class BackgroundResponse(_Response):

    def __init__(self, status, countdown=30, output=None):
        super().__init__(status, output)
        self.countdown = 30 if countdown < 30 else countdown

    @classmethod
    def skip(cls, output=None):
        return cls(ResultType.SKIP, output=output)

    @classmethod
    def reschedule(cls, countdown=30):
        return cls(ResultType.RESCHEDULE, countdown=countdown)

    @classmethod
    def slow_process_reschedule(cls, countdown=300):
        return cls(
            ResultType.RESCHEDULE,
            countdown=300 if countdown < 300 else countdown,
        )


class ProcessingResponse(BackgroundResponse):
    pass


class InteractiveResponse(_Response):
    def __init__(self, status, http_status, headers, body, output):
        super().__init__(status, output)
        self.http_status = http_status
        self.headers = headers
        self.body = body

    @property
    def data(self):
        return {
            'http_status': self.http_status,
            'headers': self.headers,
            'body': self.body,
        }

    @classmethod
    def done(cls, http_status=200, headers=None, body=None):
        return cls(ResultType.SUCCESS, http_status, headers, body, None)

    @classmethod
    def fail(cls, http_status=400, headers=None, body=None, output=None):
        return cls(ResultType.FAIL, http_status, headers, body, output)


class ValidationResponse(InteractiveResponse):
    def __init__(self, status, data, output=None):
        http_status = 200 if status == ResultType.SUCCESS else 400
        super().__init__(status, http_status, None, data, output)

    @classmethod
    def done(cls, data):
        return cls(ResultType.SUCCESS, data)

    @classmethod
    def fail(cls, data=None, output=None):
        return cls(ResultType.FAIL, data=data, output=output)


class CustomEventResponse(InteractiveResponse):
    pass


class ProductActionResponse(InteractiveResponse):
    pass


class ScheduledExecutionResponse(_Response):
    pass



import importlib
import inspect
import os
import re
import sys
from collections import deque

import yaml

from connect.eaas.core.extension import (
    AnvilExtension,
    EventsExtension,
    Extension,
    WebAppExtension,
)
from connect.eaas.core.responses import (
    CustomEventResponse,
    ProcessingResponse,
    ProductActionResponse,
    ValidationResponse,
)
from connect.eaas.core.validation.helpers import (
    get_code_context,
    get_event_definitions,
    load_project_toml_file,
)
from connect.eaas.core.validation.proto import (
    ValidationItem,
    ValidationResult,
)


def validate_pyproject_toml(project_dir, *args, **kwargs):  # noqa: CCR001
    messages = []

    data = load_project_toml_file(project_dir)
    if isinstance(data, ValidationResult):
        return data

    descriptor_file = os.path.join(project_dir, 'pyproject.toml')
    dependencies = data['tool']['poetry']['dependencies']

    if 'connect-extension-runner' in dependencies:
        messages.append(
            ValidationItem(
                message=(
                    'Extensions must depend on *connect-eaas-core* library not '
                    '*connect-extension-runner*.'
                ),
                file=descriptor_file,
            ),
        )
    elif 'connect-eaas-core' not in dependencies:
        messages.append(
            ValidationItem(
                level='ERROR',
                message='No dependency on *connect-eaas-core* has been found.',
                file=descriptor_file,
            ),
        )

    extension_dict = data['tool']['poetry'].get('plugins', {}).get('connect.eaas.ext')
    if not isinstance(extension_dict, dict):
        messages.append(
            ValidationItem(
                level='ERROR',
                message=(
                    'No extension declaration has been found.'
                    'The extension must be declared in the '
                    '*[tool.poetry.plugins."connect.eaas.ext"]* section.'
                ),
                file=descriptor_file,
            ),
        )
        return ValidationResult(items=messages, must_exit=True)

    sys.path.append(os.path.join(os.getcwd(), project_dir))
    possible_extensions = ['extension', 'webapp', 'anvil']
    extensions = {}
    for extension_type in possible_extensions:
        if extension_type in extension_dict.keys():
            package, class_name = extension_dict[extension_type].rsplit(':', 1)
            try:
                extension_module = importlib.import_module(package)
            except ImportError as err:
                messages.append(
                    ValidationItem(
                        level='ERROR',
                        message=(
                            f'The extension class *{extension_dict[extension_type]}* '
                            f'cannot be loaded: {err}.'
                        ),
                        file=descriptor_file,
                    ),
                )
                return ValidationResult(items=messages, must_exit=True)

            defined_classes = [
                member[1]
                for member in inspect.getmembers(extension_module, predicate=inspect.isclass)
            ]

            for deprecated_cls, cls_name in (
                (CustomEventResponse, 'InteractiveResponse'),
                (ProcessingResponse, 'BackgroundResponse'),
                (ProductActionResponse, 'InteractiveResponse'),
                (ValidationResponse, 'InteractiveResponse'),
            ):
                if deprecated_cls in defined_classes:
                    messages.append(
                        ValidationItem(
                            message=(
                                f'The response class *{deprecated_cls.__name__}* '
                                f'has been deprecated in favor of *{cls_name}*.'
                            ),
                            **get_code_context(extension_module, deprecated_cls.__name__),
                        ),
                    )

            extensions[extension_type] = getattr(extension_module, class_name)

    if not extensions:
        messages.append(
            ValidationItem(
                level='ERROR',
                message=(
                    'Invalid extension declaration in *[tool.poetry.plugins."connect.eaas.ext"]*: '
                    'The extension must be declared as: *"extension" = '
                    '"your_package.extension:YourExtension"* '
                    'for Fulfillment automation or Hub integration. '
                    'For Multi account installation must be '
                    'declared at least one the following: *"extension" = '
                    '"your_package.events:YourEventsExtension"*, '
                    '*"webapp" = "your_package.webapp:YourWebAppExtension"*, '
                    '*"anvil" = "your_package.anvil:YourAnvilExtension"*.'
                ),
                file=descriptor_file,
            ),
        )
        return ValidationResult(items=messages, must_exit=True)

    return ValidationResult(items=messages, context={'extension_classes': extensions})


def validate_extension_class(context, *args, **kwargs):  # noqa: CCR001
    messages = []
    class_mapping = {
        'extension': 'connect.eaas.core.extension.[Events]Extension',
        'webapp': 'connect.eaas.core.extension.WebAppExtension',
        'anvil': 'connect.eaas.core.extension.AnvilExtension',
    }
    ext_class = None
    extension_json_file = None

    for extension_type, extension_class in context['extension_classes'].items():
        extension_class_file = inspect.getsourcefile(extension_class)

        if (
            extension_type == 'extension'
            and not issubclass(extension_class, (Extension, EventsExtension))
            or extension_type == 'webapp' and not issubclass(extension_class, WebAppExtension)
            or extension_type == 'anvil' and not issubclass(extension_class, AnvilExtension)
        ):
            messages.append(
                ValidationItem(
                    level='ERROR',
                    message=(
                        f'The extension class *{extension_class.__name__}* '
                        f'is not a subclass of *{class_mapping[extension_type]}*.'
                    ),
                    file=extension_class_file,
                ),
            )
            return ValidationResult(items=messages, must_exit=True)

        if not extension_json_file:  # pragma: no branch
            extension_json_file = os.path.join(
                os.path.dirname(extension_class_file),
                'extension.json',
            )
            ext_class = extension_class

    try:
        descriptor = ext_class.get_descriptor()
    except FileNotFoundError:
        messages.append(
            ValidationItem(
                level='ERROR',
                message='The extension descriptor *extension.json* cannot be loaded.',
                file=extension_json_file,
            ),
        )
        return ValidationResult(items=messages, must_exit=True)

    for description in ['variables', 'capabilities', 'schedulables']:
        if description in descriptor:
            messages.append(
                ValidationItem(
                    message=(
                        f'Extension {description} must be declared using the '
                        f'*connect.eaas.core.decorators.'
                        f'{description if description != "schedulables" else "event"}* decorator.'
                    ),
                    file=extension_json_file,
                ),
            )

    return ValidationResult(items=messages, context={'descriptor': descriptor})


def validate_events(config, context, *args, **kwargs):
    messages = []

    extension_class = context['extension_classes'].get('extension')

    if not extension_class:
        return ValidationResult()

    definitions = {
        definition['type']: definition for definition in get_event_definitions(
            config.active.client,
        )
    }
    events = extension_class.get_events()
    for event in events:
        method = getattr(extension_class, event['method'])
        if event['event_type'] not in definitions:
            messages.append(
                ValidationItem(
                    level='ERROR',
                    message=f'The event type *{event["event_type"]}* is not valid.',
                    **get_code_context(method, '@event'),
                ),
            )
            continue

        statuses = definitions[event['event_type']]['object_statuses']
        if statuses:
            invalid_statuses = set(event['statuses']) - set(statuses)
        else:
            invalid_statuses = set(event['statuses'] or [])
        if invalid_statuses:
            messages.append(
                ValidationItem(
                    level='ERROR',
                    message=(
                        f'The status/es *{", ".join(invalid_statuses)}* are invalid '
                        f'for the event *{event["event_type"]}*.'
                    ),
                    **get_code_context(method, '@event'),
                ),
            )

        signature = inspect.signature(method)
        if len(signature.parameters) != 2:
            sig_str = f'{event["method"]}({", ".join(signature.parameters)})'

            messages.append(
                ValidationItem(
                    level='ERROR',
                    message=(
                        f'The handler for the event *{event["event_type"]}* '
                        f'has an invalid signature: *{sig_str}*'
                    ),
                    **get_code_context(method, sig_str),
                ),
            )
    return ValidationResult(items=messages)


def validate_schedulables(context, *args, **kwargs):
    messages = []

    extension_class = context['extension_classes'].get('extension')

    if not extension_class:
        return ValidationResult()

    schedulables = extension_class.get_schedulables()
    for schedulable in schedulables:
        method = getattr(extension_class, schedulable['method'])
        signature = inspect.signature(method)
        if len(signature.parameters) != 2:
            sig_str = f'{schedulable["method"]}({", ".join(signature.parameters)})'

            messages.append(
                ValidationItem(
                    level='ERROR',
                    message=(
                        f'The schedulable method *{schedulable["method"]}* '
                        f'has an invalid signature: *{sig_str}*'
                    ),
                    **get_code_context(method, sig_str),
                ),
            )
    return ValidationResult(items=messages)


def validate_variables(context, *args, **kwargs):  # noqa: CCR001

    messages = []

    for _, extension_class in context['extension_classes'].items():

        variables = extension_class.get_variables()
        variable_name_pattern = r'^[A-Za-z](?:[A-Za-z0-9_\-.]+)*$'
        variable_name_regex = re.compile(variable_name_pattern)

        names = []

        for variable in variables:
            if 'name' not in variable:
                messages.append(
                    ValidationItem(
                        level='ERROR',
                        message='Invalid variable declaration: the *name* attribute is mandatory.',
                        **get_code_context(extension_class, '@variables'),
                    ),
                )
                continue

            if variable["name"] in names:
                messages.append(
                    ValidationItem(
                        level='ERROR',
                        message=(
                            f'Duplicate variable name: the variable with name *{variable["name"]}* '
                            'has already been declared.'
                        ),
                        **get_code_context(extension_class, '@variables'),
                    ),
                )

            names.append(variable["name"])

            if not variable_name_regex.match(variable['name']):
                messages.append(
                    ValidationItem(
                        level='ERROR',
                        message=(
                            f'Invalid variable name: the value *{variable["name"]}* '
                            f'does not match the pattern *{variable_name_pattern}*.'
                        ),
                        **get_code_context(extension_class, '@variables'),
                    ),
                )
            if 'initial_value' in variable and not isinstance(variable['initial_value'], str):
                messages.append(
                    ValidationItem(
                        level='ERROR',
                        message=(
                            f'Invalid *initial_value* attribute for variable *{variable["name"]}*: '
                            f'must be a non-null string not *{type(variable["initial_value"])}*.'
                        ),
                        **get_code_context(extension_class, '@variables'),
                    ),
                )

            if 'secure' in variable and not isinstance(variable['secure'], bool):
                messages.append(
                    ValidationItem(
                        level='ERROR',
                        message=(
                            f'Invalid *secure* attribute for variable *{variable["name"]}*: '
                            f'must be a boolean not *{type(variable["secure"])}*.'
                        ),
                        **get_code_context(extension_class, '@variables'),
                    ),
                )

    return ValidationResult(items=messages)


def validate_webapp_extension(context, *args, **kwargs):  # noqa: CCR001

    messages = []

    if 'webapp' not in context['extension_classes']:
        return ValidationResult()

    extension_class = context['extension_classes']['webapp']
    extension_class_file = inspect.getsourcefile(extension_class)

    if not inspect.getsource(extension_class).strip().startswith('@web_app(router)'):
        messages.append(
            ValidationItem(
                level='ERROR',
                message='The Web app extension class must be wrapped in *@web_app(router)*.',
                file=extension_class_file,
            ),
        )
        return ValidationResult(items=messages, must_exit=True)

    has_router_function = False
    for _, value in inspect.getmembers(extension_class):
        if (
                inspect.isfunction(value)
                and inspect.getsource(value).strip().startswith('@router.')
        ):
            has_router_function = True
            break

    if not has_router_function:
        messages.append(
            ValidationItem(
                level='ERROR',
                message=(
                    'The Web app extension class must contain at least one router '
                    'implementation function wrapped in *@router.your_method("/your_path")*.'
                ),
                file=extension_class_file,
            ),
        )
        return ValidationResult(items=messages, must_exit=True)

    if 'ui' in context['descriptor']:
        extension_json_file = os.path.join(os.path.dirname(extension_class_file), 'extension.json')

        ui_items = deque()
        missed_files = []
        for _, value in context['descriptor']['ui'].items():
            ui_items.append(value)

        while ui_items:
            ui_item = ui_items.pop()
            try:
                url = ui_item['url']
            except KeyError:
                messages.append(
                    ValidationItem(
                        level='ERROR',
                        message=(
                            'The extension descriptor contains incorrect ui item'
                            f'*{ui_item.get("label")}*, url is not presented.'
                        ),
                        file=extension_json_file,
                    ),
                )
                return ValidationResult(items=messages, must_exit=True)

            path = os.path.join(
                os.path.dirname(extension_class_file),
                'static_root',
                url.split('/')[-1],
            )
            if not os.path.exists(path):
                missed_files.append(url)

            for child in ui_item.get('children', []):
                ui_items.append(child)

        if missed_files:
            messages.append(
                ValidationItem(
                    level='ERROR',
                    message=(
                        'The extension descriptor contains missing static files: '
                        f'{", ".join(missed_files)}.'
                    ),
                    file=extension_json_file,
                ),
            )
            return ValidationResult(items=messages, must_exit=True)

    return ValidationResult(items=messages)


def validate_anvil_extension(context, *args, **kwargs):

    messages = []

    if 'anvil' not in context['extension_classes']:
        return ValidationResult()

    extension_class = context['extension_classes']['anvil']
    anvil_key_var = extension_class.get_anvil_key_variable()

    if anvil_key_var:
        variable_name_pattern = r'^[A-Za-z](?:[A-Za-z0-9_\-.]+)*$'
        variable_name_regex = re.compile(variable_name_pattern)

        if not variable_name_regex.match(anvil_key_var):
            messages.append(
                ValidationItem(
                    level='ERROR',
                    message=(
                        f'Invalid Anvil key variable name: the value *{anvil_key_var}* '
                        f'does not match the pattern *{variable_name_pattern}*.'
                    ),
                    file=inspect.getsourcefile(extension_class),
                ),
            )

    return ValidationResult(items=messages)


def validate_docker_compose_yml(project_dir, context, *args, **kwargs):
    messages = []
    compose_file = os.path.join(project_dir, 'docker-compose.yml')
    if not os.path.isfile(compose_file):
        messages.append(
            ValidationItem(
                message=(
                    f'The directory *{project_dir}* does not look like an extension project '
                    'directory, the file *docker-compose.yml* is not present.'
                ),
                file=compose_file,
            ),
        )
        return ValidationResult(items=messages)
    try:
        data = yaml.safe_load(open(compose_file, 'r'))
    except yaml.YAMLError:
        messages.append(
            ValidationItem(
                level='ERROR',
                message='The file *docker-compose.yml* is not valid.',
                file=compose_file,
            ),
        )
        return ValidationResult(items=messages)

    runner_image = f'cloudblueconnect/connect-extension-runner:{context["runner_version"]}'

    for service in data['services']:
        image = data['services'][service].get('image')
        if image != runner_image:
            messages.append(
                ValidationItem(
                    level='ERROR',
                    message=(
                        f'Invalid image for service *{service}*: '
                        f'expected *{runner_image}* got *{image}*.'
                    ),
                    file=compose_file,
                ),
            )
    return ValidationResult(items=messages)


extension_validators = [
    validate_pyproject_toml,
    validate_docker_compose_yml,
    validate_extension_class,
    validate_events,
    validate_variables,
    validate_schedulables,
    validate_webapp_extension,
    validate_anvil_extension,
]



from typing import Any, List, Literal, Optional, Union

from pydantic import BaseModel, Field


class MessageType:
    SETUP_REQUEST = 'setup_request'
    SETUP_RESPONSE = 'setup_response'
    WEB_SETUP_REQUEST = 'web_setup_request'
    WEB_SETUP_RESPONSE = 'web_setup_response'
    WEB_TASK = 'web_task'
    TASK = 'task'
    SHUTDOWN = 'shutdown'
    # delete after stop using version 1
    CAPABILITIES = 'capabilities'
    CONFIGURATION = 'configuration'


class TaskOptions(BaseModel):
    task_id: str
    task_category: str
    correlation_id: Optional[str]
    reply_to: Optional[str]
    api_key: Optional[str]
    installation_id: Optional[str]
    connect_correlation_id: Optional[str]


class TaskOutput(BaseModel):
    result: str
    data: Optional[Any]
    countdown: int = 0
    runtime: float = 0.0
    message: Optional[str]


class TaskInput(BaseModel):
    event_type: str
    object_id: str
    data: Optional[Any]


class Task(BaseModel):
    options: TaskOptions
    input: TaskInput
    output: Optional[TaskOutput]
    model_type: Literal['task'] = 'task'


class LogMeta(BaseModel):
    account_id: Optional[str]
    account_name: Optional[str]
    service_id: Optional[str]
    # delete after stop using version 1
    products: Optional[List[str]]
    hub_id: Optional[str]


class Logging(BaseModel):
    logging_api_key: Optional[str]
    log_level: Optional[str]
    runner_log_level: Optional[str]
    meta: Optional[LogMeta]


class EventDefinition(BaseModel):
    event_type: str
    api_resource_endpoint: str
    api_collection_endpoint: str
    api_collection_filter: str


class SetupResponse(BaseModel):
    variables: Optional[dict]
    # delete after stop using version 1
    environment_type: Optional[str]
    logging: Optional[Logging]
    event_definitions: Optional[List[EventDefinition]]
    model_type: Literal['setup_response'] = 'setup_response'


class Schedulable(BaseModel):
    method: str
    name: str
    description: str


class Repository(BaseModel):
    readme_url: Optional[str]
    changelog_url: Optional[str]


class SetupRequest(BaseModel):
    # for version 1 'capabilities' renaming to 'event_subscriptions'
    event_subscriptions: Optional[dict]
    ui_modules: Optional[dict]
    icon: Optional[str]
    variables: Optional[list]
    schedulables: Optional[List[Schedulable]]
    repository: Optional[Repository]
    runner_version: Optional[str]
    model_type: Literal['setup_request'] = 'setup_request'


class HttpResponse(BaseModel):
    status: int
    headers: dict
    content: Optional[Any]


class HttpRequest(BaseModel):
    method: str
    url: str
    headers: dict
    content: Optional[Any]


class WebTaskOptions(BaseModel):
    correlation_id: str
    reply_to: str
    api_key: Optional[str]
    installation_id: Optional[str]
    connect_correlation_id: Optional[str]
    user_id: Optional[str]
    account_id: Optional[str]
    account_role: Optional[str]
    call_type: Optional[Literal['admin', 'user']]
    call_source: Optional[Literal['ui', 'api']]


class WebTask(BaseModel):
    options: WebTaskOptions
    request: Optional[HttpRequest]
    response: Optional[HttpResponse]
    model_type: Literal['web_task'] = 'web_task'


class Message(BaseModel):
    version: Literal[1, 2] = 1
    message_type: str
    data: Union[
        WebTask, Task,
        SetupRequest, SetupResponse,
        None,
    ] = Field(discriminator='model_type')

    def serialize(self, protocol_version=2):
        if protocol_version == 2:
            return self.dict()

        if self.message_type == MessageType.SETUP_REQUEST:
            return {
                'message_type': MessageType.CAPABILITIES,
                'data': {
                    'capabilities': self.data.event_subscriptions,
                    'variables': self.data.variables,
                    'schedulables': [
                        {
                            'method': schedulable.method,
                            'name': schedulable.name,
                            'description': schedulable.description,
                        }
                        for schedulable in (self.data.schedulables or [])
                    ],
                    'readme_url': self.data.repository.readme_url,
                    'changelog_url': self.data.repository.changelog_url,
                    'runner_version': self.data.runner_version,
                },
            }
        if self.message_type == MessageType.SETUP_RESPONSE:
            if not self.data.logging.meta:
                self.data.logging.meta = LogMeta()
            return {
                'message_type': MessageType.CONFIGURATION,
                'data': {
                    'configuration': self.data.variables,
                    'environment_type': self.data.environment_type,
                    'logging_api_key': self.data.logging.logging_api_key,
                    'log_level': self.data.logging.log_level,
                    'runner_log_level': self.data.logging.runner_log_level,
                    'account_id': self.data.logging.meta.account_id,
                    'account_name': self.data.logging.meta.account_name,
                    'service_id': self.data.logging.meta.service_id,
                    'product_id': (
                        ','.join(self.data.logging.meta.products)
                        if self.data.logging.meta.products
                        else None
                    ),
                    'hub_id': self.data.logging.meta.hub_id,
                },
            }

        if self.message_type == MessageType.TASK:
            return {
                'message_type': self.message_type,
                'data': {
                    'task_id': self.data.options.task_id,
                    'task_category': self.data.options.task_category,
                    'correlation_id': self.data.options.correlation_id,
                    'reply_to': self.data.options.reply_to,
                    'task_type': self.data.input.event_type,
                    'object_id': self.data.input.object_id,
                    'data': self.data.input.data,
                },
            }
        return {
            'message_type': self.message_type,
        }

    @classmethod
    def deserialize(cls, raw):
        version = raw.get('version', 1)

        message_type = raw['message_type']
        raw_data = raw.get('data')

        if version == 2:
            return cls(**raw)

        if message_type == MessageType.CAPABILITIES:
            return cls(
                version=version,
                message_type=MessageType.SETUP_REQUEST,
                data=SetupRequest(
                    event_subscriptions=raw_data['capabilities'],
                    runner_version=raw_data.get('runner_version'),
                    variables=raw_data.get('variables'),
                    schedulables=[
                        Schedulable(**schedulable)
                        for schedulable in (raw_data.get('schedulables', []) or [])
                    ],
                    repository=Repository(
                        readme_url=raw_data.get('readme_url'),
                        changelog_url=raw_data.get('changelog_url'),
                    ),
                ),
            )
        if message_type == MessageType.CONFIGURATION:
            return cls(
                version=version,
                message_type=MessageType.SETUP_RESPONSE,
                data=SetupResponse(
                    environment_type=raw_data.get('environment_type'),
                    variables=raw_data.get('configuration'),
                    logging=Logging(
                        logging_api_key=raw_data.get('logging_api_key'),
                        log_level=raw_data.get('log_level'),
                        runner_log_level=raw_data.get('runner_log_level'),
                        meta=LogMeta(
                            account_id=raw_data.get('account_id'),
                            account_name=raw_data.get('account_name'),
                            service_id=raw_data.get('service_id'),
                            products=[raw_data['product_id']] if 'product_id' in raw else None,
                            hub_id=raw_data.get('hub_id'),
                        ),
                    ),
                ),
            )
        if message_type == MessageType.TASK:
            return cls(
                version=version,
                message_type=message_type,
                data=Task(
                    options=TaskOptions(
                        task_id=raw_data['task_id'],
                        task_category=raw_data['task_category'],
                        correlation_id=raw_data.get('correlation_id'),
                        reply_to=raw_data.get('reply_to'),
                    ),
                    input=TaskInput(
                        event_type=raw_data['task_type'],
                        object_id=raw_data['object_id'],
                    ),
                    output=TaskOutput(
                        result=raw_data['result'],
                        data=raw_data.get('data'),
                        countdown=raw_data['countdown'],
                        runtime=raw_data.get('runtime'),
                        message=raw_data.get('output'),
                    ),
                ),
            )

        return cls(version=version, message_type=message_type)



import functools
import inspect
import json
import os

import anvil.server
import pkg_resources
from fastapi import APIRouter


from connect.eaas.core.constants import (
    ACCOUNT_SETTINGS_PAGE_ATTR_NAME,
    ADMIN_PAGES_ATTR_NAME,
    ANVIL_CALLABLE_ATTR_NAME,
    ANVIL_KEY_VAR_ATTR_NAME,
    EVENT_INFO_ATTR_NAME,
    GUEST_ENDPOINT_ATTR_NAME,
    MODULE_PAGES_ATTR_NAME,
    SCHEDULABLE_INFO_ATTR_NAME,
    VARIABLES_INFO_ATTR_NAME,
)
from connect.eaas.core.decorators import router


class ExtensionBase:

    @classmethod
    def get_descriptor(cls):  # pragma: no cover
        return json.load(
            pkg_resources.resource_stream(
                cls.__module__,
                'extension.json',
            ),
        )

    @classmethod
    def get_variables(cls):
        return getattr(cls, VARIABLES_INFO_ATTR_NAME, [])


class EventsExtension(ExtensionBase):
    def __init__(self, client, logger, config, installation_client=None, installation=None):
        self.client = client
        self.logger = logger
        self.config = config
        self.installation_client = installation_client
        self.installation = installation

    @classmethod
    def get_events(cls):
        return cls._get_methods_info(EVENT_INFO_ATTR_NAME)

    @classmethod
    def get_schedulables(cls):
        return cls._get_methods_info(SCHEDULABLE_INFO_ATTR_NAME)

    @classmethod
    def _get_methods_info(cls, attr_name):
        info = []
        members = inspect.getmembers(cls)
        for _, value in members:
            if not (inspect.isfunction(value) or inspect.iscoroutinefunction(value)):
                continue
            if hasattr(value, attr_name):
                info.append(getattr(value, attr_name))
        return info


class Extension(EventsExtension):
    pass


class WebAppExtension(ExtensionBase):

    @classmethod
    def get_static_root(cls):
        static_root = os.path.abspath(
            os.path.join(
                os.path.dirname(inspect.getfile(cls)),
                'static_root',
            ),
        )
        if os.path.exists(static_root) and os.path.isdir(static_root):
            return static_root
        return None

    @classmethod
    def get_routers(cls):
        auth = APIRouter()
        no_auth = APIRouter()
        for route in router.routes:
            if getattr(route.endpoint, GUEST_ENDPOINT_ATTR_NAME, False):
                no_auth.routes.append(route)
            else:
                auth.routes.append(route)
        return auth, no_auth

    @classmethod
    def get_ui_modules(cls):
        ui_modules = {}
        if hasattr(cls, ACCOUNT_SETTINGS_PAGE_ATTR_NAME):
            ui_modules['settings'] = getattr(cls, ACCOUNT_SETTINGS_PAGE_ATTR_NAME)

        if hasattr(cls, MODULE_PAGES_ATTR_NAME):
            ui_modules['modules'] = getattr(cls, MODULE_PAGES_ATTR_NAME)

        if hasattr(cls, ADMIN_PAGES_ATTR_NAME):
            ui_modules['admins'] = getattr(cls, ADMIN_PAGES_ATTR_NAME)

        return ui_modules


def _invoke(method, **kwargs):
    return method(**kwargs)


class AnvilExtension(ExtensionBase):

    def __init__(self, client, logger, config, installation_client=None, installation=None):
        self.client = client
        self.logger = logger
        self.config = config
        self.installation_client = installation_client
        self.installation = installation

    @classmethod
    def get_anvil_key_variable(cls):
        return getattr(cls, ANVIL_KEY_VAR_ATTR_NAME, [])

    def setup_anvil_callables(self):
        members = inspect.getmembers(self)
        for _, value in members:
            if not inspect.ismethod(value):
                continue

            if getattr(value, ANVIL_CALLABLE_ATTR_NAME, False):
                fn = functools.partial(_invoke, value)
                fn.__name__ = value.__name__
                anvil.server.callable(fn)



import json
import logging

from fastapi import Depends, Header

from connect.eaas.core.inject.models import Context
from connect.eaas.core.logging import ExtensionLogHandler

_LOGGING_HANDLER = None


def get_call_context(
    x_connect_installation_id: str = Header(),
    x_connect_user_id: str = Header(),
    x_connect_account_id: str = Header(),
    x_connect_account_role: str = Header(),
    x_connect_call_source: str = Header(),
    x_connect_call_type: str = Header(),
) -> Context:
    return Context(
        installation_id=x_connect_installation_id,
        user_id=x_connect_user_id,
        account_id=x_connect_account_id,
        account_role=x_connect_account_role,
        call_source=x_connect_call_source,
        call_type=x_connect_call_type,
    )


def get_logger(
    x_connect_logging_api_key: str = Header(None),
    x_connect_logging_metadata: str = Header('{}'),
    x_connect_logging_level: str = Header('INFO'),
    context: Context = Depends(get_call_context),
):
    global _LOGGING_HANDLER

    logger = logging.getLogger('eaas.webapp')
    if _LOGGING_HANDLER is None and x_connect_logging_api_key is not None:
        _LOGGING_HANDLER = ExtensionLogHandler(
            x_connect_logging_api_key,
            default_extra_fields=json.loads(x_connect_logging_metadata),
        )
        logger.addHandler(_LOGGING_HANDLER)

    logger.setLevel(
        getattr(logging, x_connect_logging_level),
    )
    return logging.LoggerAdapter(
        logger,
        context.dict(),
    )



import os
from logging import Logger

from fastapi import Depends, Header

from connect.client import ConnectClient
from connect.eaas.core.inject.common import get_logger
from connect.eaas.core.logging import RequestLogger


def get_installation_client(
    logger: Logger = Depends(get_logger),
    x_connect_installation_api_key: str = Header(),
    x_connect_api_gateway_url: str = Header(),
    x_connect_user_agent: str = Header(),
    x_connect_correlation_id: str = Header(None),
):

    default_headers = {
        'User-Agent': x_connect_user_agent,
    }
    if x_connect_correlation_id:
        default_headers['traceparent'] = x_connect_correlation_id

    return ConnectClient(
        x_connect_installation_api_key,
        endpoint=x_connect_api_gateway_url,
        use_specs=False,
        default_headers=default_headers,
        logger=RequestLogger(logger),
    )


def get_extension_client(
    logger: Logger = Depends(get_logger),
    x_connect_api_gateway_url: str = Header(),
    x_connect_user_agent: str = Header(),
    x_connect_correlation_id: str = Header(None),
):
    default_headers = {
        'User-Agent': x_connect_user_agent,
    }
    if x_connect_correlation_id:
        default_headers['traceparent'] = x_connect_correlation_id

    return ConnectClient(
        os.getenv('API_KEY'),
        endpoint=x_connect_api_gateway_url,
        use_specs=False,
        default_headers=default_headers,
        logger=RequestLogger(logger),
    )


def get_installation(
    client: ConnectClient = Depends(get_installation_client),
    x_connect_installation_id: str = Header(),
):
    return client('devops').installations[x_connect_installation_id].get()


def get_environment(
    client: ConnectClient = Depends(get_extension_client),
    x_connect_extension_id: str = Header(),
    x_connect_environment_id: str = Header(),
):
    extension = client('devops').services[x_connect_extension_id]
    return {
        variable['name']: variable['value']
        for variable in extension.environments[x_connect_environment_id].variables.all()
    }



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['connect',
 'connect.eaas.core',
 'connect.eaas.core.inject',
 'connect.eaas.core.validation']

package_data = \
{'': ['*']}

install_requires = \
['anvil-uplink>=0.4.0,<0.5.0',
 'connect-openapi-client>=25.10',
 'fastapi-utils>=0.2.1,<0.3.0',
 'fastapi>=0.78.0,<0.79.0',
 'logzio-python-handler>=3.1.1,<4.0.0',
 'pydantic>=1.9.0,<2.0.0']

setup_kwargs = {
    'name': 'connect-eaas-core',
    'version': '4.0.0',
    'description': 'Connect Eaas Core',
    'long_description': '',
    'author': 'CloudBlue LLC',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://connect.cloudblue.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles
from starlette.testclient import TestClient


class WebAppTestClient(TestClient):

    def __init__(self, webapp, default_headers=None):
        self._webapp_class = webapp
        self.app = self.get_application()

        super().__init__(app=self.app)

        if default_headers:
            self.headers = default_headers

    def get_application(self):
        app = FastAPI()

        auth_router, no_auth_router = self._webapp_class.get_routers()
        app.include_router(auth_router, prefix='/api')
        app.include_router(no_auth_router, prefix='/guest')

        static_root = self._webapp_class.get_static_root()
        if static_root:
            app.mount('/static', StaticFiles(directory=static_root), name='static')

        return app



from fastapi_utils.cbv import cbv
from fastapi_utils.inferring_router import InferringRouter

from connect.eaas.core.constants import (
    ACCOUNT_SETTINGS_PAGE_ATTR_NAME,
    ADMIN_PAGES_ATTR_NAME,
    ANVIL_CALLABLE_ATTR_NAME,
    ANVIL_KEY_VAR_ATTR_NAME,
    EVENT_INFO_ATTR_NAME,
    GUEST_ENDPOINT_ATTR_NAME,
    MODULE_PAGES_ATTR_NAME,
    SCHEDULABLE_INFO_ATTR_NAME,
    VARIABLES_INFO_ATTR_NAME,
)


def event(event_type, statuses=None):
    def wrapper(func):
        setattr(
            func,
            EVENT_INFO_ATTR_NAME,
            {
                'method': func.__name__,
                'event_type': event_type,
                'statuses': statuses,
            },
        )
        return func
    return wrapper


def schedulable(name, description):
    def wrapper(func):
        setattr(
            func,
            SCHEDULABLE_INFO_ATTR_NAME,
            {
                'method': func.__name__,
                'name': name,
                'description': description,
            },
        )
        return func
    return wrapper


def variables(variables):
    def wrapper(cls):
        if hasattr(cls, VARIABLES_INFO_ATTR_NAME):
            declared_vars = getattr(cls, VARIABLES_INFO_ATTR_NAME)
            for variable in variables:
                name = variable['name']
                if len(list(filter(lambda x: x['name'] == name, declared_vars))) == 0:
                    declared_vars.append(variable)
            return cls
        setattr(cls, VARIABLES_INFO_ATTR_NAME, variables)
        return cls
    return wrapper


def anvil_key_variable(name):
    def wrapper(cls):
        setattr(cls, ANVIL_KEY_VAR_ATTR_NAME, name)
        variables = []
        if not hasattr(cls, VARIABLES_INFO_ATTR_NAME):
            setattr(cls, VARIABLES_INFO_ATTR_NAME, variables)
        else:
            variables = getattr(cls, VARIABLES_INFO_ATTR_NAME)
        if len(list(filter(lambda x: x['name'] == name, variables))) == 0:
            variables.append({'name': name, 'initial_value': 'changeme!', 'secure': True})
        return cls
    return wrapper


def anvil_callable():
    def wrapper(func):
        setattr(func, ANVIL_CALLABLE_ATTR_NAME, True)
        return func
    return wrapper


def guest():
    def wrapper(func):
        setattr(func, GUEST_ENDPOINT_ATTR_NAME, True)
        return func
    return wrapper


def account_settings_page(label, url):
    def wrapper(cls):
        setattr(cls, ACCOUNT_SETTINGS_PAGE_ATTR_NAME, {'label': label, 'url': url})
        return cls
    return wrapper


def module_pages(label, url, children=None):
    def wrapper(cls):
        data = {
            'label': label,
            'url': url,
        }
        if children:
            data['children'] = children
        setattr(cls, MODULE_PAGES_ATTR_NAME, data)
        return cls
    return wrapper


def admin_pages(pages):
    def wrapper(cls):
        setattr(cls, ADMIN_PAGES_ATTR_NAME, pages)
        return cls
    return wrapper


router = InferringRouter()
web_app = cbv



VARIABLES_INFO_ATTR_NAME = '_eaas_variable_info'
EVENT_INFO_ATTR_NAME = '_eaas_event_info'
SCHEDULABLE_INFO_ATTR_NAME = '_eaas_schedulable_info'
ANVIL_KEY_VAR_ATTR_NAME = '_eaas_anvil_key_var'
ANVIL_CALLABLE_ATTR_NAME = '_eaas_anvil_callable'
GUEST_ENDPOINT_ATTR_NAME = '_eaas_guest_endpoint'
ACCOUNT_SETTINGS_PAGE_ATTR_NAME = '_eaas_account_settings_page'
MODULE_PAGES_ATTR_NAME = '_eaas_module_pages'
ADMIN_PAGES_ATTR_NAME = '_eaas_admin_pages'



import inspect
import os

import toml

from connect.eaas.core.validation.proto import ValidationItem, ValidationResult


def get_event_definitions(client):
    return list(client('devops').event_definitions.all())


def get_code_context(obj, pattern):
    source_lines = inspect.getsourcelines(obj)
    file = inspect.getsourcefile(obj)
    start_line = source_lines[1]
    lineno = source_lines[1]
    code = ''.join(source_lines[0])

    for idx, line in enumerate(source_lines[0]):  # pragma: no branch
        if pattern in line:
            lineno += idx
            break

    if inspect.ismodule(obj):
        code = ''.join(code.splitlines(keepends=True)[0:lineno + 3])

    return {
        'file': file,
        'start_line': start_line,
        'lineno': lineno,
        'code': code,
    }


def load_project_toml_file(path):
    descriptor_file = os.path.join(path, 'pyproject.toml')
    if not os.path.isfile(descriptor_file):
        return ValidationResult(
            items=[
                ValidationItem(
                    level='ERROR',
                    message=(
                        'The mandatory *pyproject.toml* project '
                        f'descriptor file is not present in the folder {path}.'
                    ),
                    file=descriptor_file,
                ),
            ],
            must_exit=True,
        )
    try:
        return toml.load(descriptor_file)
    except toml.TomlDecodeError:
        return ValidationResult(
            items=[
                ValidationItem(
                    level='ERROR',
                    message=(
                        'The project descriptor file *pyproject.toml* is '
                        'not a valid toml file.'
                    ),
                    file=descriptor_file,
                ),
            ],
            must_exit=True,
        )



from typing import List, Literal, Optional

from pydantic import BaseModel


class ValidationItem(BaseModel):
    level: Literal['WARNING', 'ERROR'] = 'WARNING'
    message: str
    file: Optional[str]
    start_line: Optional[int]
    lineno: Optional[int]
    code: Optional[str]


class ValidationResult(BaseModel):
    items: List[ValidationItem] = []
    must_exit: bool = False
    context: Optional[dict]
